"""
SCV:
    I/O tool kit.
"""
import os
import csv
import datetime, time
import json
import re
import pyodbc
from azure.storage import CloudStorageAccount
import azure.cosmos.documents as documents
import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.errors as cosmos_errors

#region Local I/O
def read_json(file_path):
    with open(file_path) as f:
        s = f.read()
        jstr = json.loads(s)
        return jstr

def read_json_default(file_name='x'):
    file_path = os.path.join(os.path.expanduser('~'), 'Downloads', \
        '{}.json'.format(file_name))
    return read_json(file_path)

def beautify_json_default(file_name='x'):
    file_path = os.path.join(os.path.expanduser('~'), 'Downloads', \
        '{}.json'.format(file_name))
    j_str = read_json(file_path)
    if 'body' in j_str:
        j_str = j_str['body']
    write_json_default(j_str, file_name)
    write_json(j_str, mkdownloadspath('resources', 'debug', '{}.json'\
        .format(file_name)))
    write_json(j_str, mkdownloadspath('resources', 'debug', '{}_{}.json'\
        .format(file_name, datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))))
    return j_str

def read_lines(file_path):
    with open(file_path) as f:
        lines = f.readlines()
        return lines

def read_csv_list(file_path):
    with open(file_path) as f:
        return f.readlines()

def read_csv_list_default(file_name='x'):
    file_path = os.path.join(os.path.expanduser('~'), 'Downloads', \
        '{}.csv'.format(file_name))
    return read_csv_list(file_path)

def read_csv_dict_list(file_path):
    """
    Return list of dict.
    """
    with open(file_path, encoding='utf-8-sig') as f:
        row_list = []
        lines = list(csv.reader(f, skipinitialspace=True))
        row_list = []
        headers = [x.strip() for x in lines[0]]
        for i in range(1, len(lines)):
            row_dict = {}
            row_values = lines[i]
            for j in range(0, len(row_values)):
                if not row_values[j] or row_values[j].lower() == 'null':
                    continue
                row_dict[headers[j]] = row_values[j]
            row_list.append(row_dict)
        return row_list

def read_csv_dict_list_default(file_name='x'):
    file_path = os.path.join(os.path.expanduser('~'), 'Downloads', \
        '{}.csv'.format(file_name))
    return read_csv_dict_list(file_path)

def write_json(j_str, file_path):
    with open(file_path, 'w') as f:
        f.write(json.dumps(j_str, default=_datetime_handler, \
                sort_keys=False, indent=2))

def write_json_default(j_str, file_name='x'):
    file_path = os.path.join(os.path.expanduser('~'), 'Downloads', \
        '{}.json'.format(file_name))
    write_json(j_str, file_path)

def write_lines(lines, file_path):
    with open(file_path, 'w') as f:
        f.write('')
    with open(file_path, 'a') as f:
        for line in lines:
            f.write(line)

def append_line(line, file_path):
    with open(file_path, 'a') as f:
        f.write(line)
#endregion

#region SQL I/O
def query_rows(connection_string, query, \
        query_size=100000):
    """Return dict list."""
    with pyodbc.connect(connection_string) as cnxn:
        cursor = cnxn.cursor()
        cursor.execute(query)
        rows = cursor.fetchmany(query_size)
        full_query_result = {'description': cursor.description, 'rows': rows}
        return _extract_dict_list_from_sql_query_result(full_query_result)
    
def query_cell(connection_string, query):
    return execute_query(connection_string, query)

def execute_query(connection_string, query):
    with pyodbc.connect(connection_string) as cnxn:
        cursor = cnxn.cursor()
        cursor.execute(query)
        res = cursor.fetchone()
        cursor.commit()
        return res[0]
#endregion

#region AzureBlob I/O
def get_blob(account_name, account_key, \
        container_name, blob_name):
    """Return json format."""
    account = CloudStorageAccount(account_name, account_key)
    blockblob_service = account.create_block_blob_service()
    blob_object = blockblob_service.get_blob_to_text(container_name, blob_name)
    return json.loads(blob_object.content)

def upload_to_blob(account_name, account_key, \
        container_name, blob_name, file_path):
    """Upload file (path) to blob."""
    account = CloudStorageAccount(account_name, account_key)
    blockblob_service = account.create_block_blob_service()
    blockblob_service.create_container(container_name)
    blockblob_service.create_blob_from_path(container_name, blob_name, file_path)
#endregion

#region Cosmos I/O
def insert_doc(cosmos_host, cosmos_key, \
        db_name, collection_name, doc):
    db_link = 'dbs/{}'.format(db_name)
    collection_link = '{}/colls/{}'.format(db_link, collection_name)
    with IDisposable(cosmos_client.CosmosClient(cosmos_host, \
            {'masterKey': cosmos_key})) as client:
        collection = _get_or_create_collection_if_not_exist(\
            client, db_name, collection_name)
        if isinstance(doc, (list,)):
            for d in doc:
                collection.CreateItem(collection_link, d)
        else:
            collection.CreateItem(collection_link, doc)

def update_doc(cosmos_host, cosmos_key, \
        db_name, collection_name, doc_id, new_doc):
    db_link = 'dbs/{}'.format(db_name)
    collection_link = '{}/colls/{}'.format(db_link, collection_name)
    with IDisposable(cosmos_client.CosmosClient(cosmos_host, \
            {'masterKey': cosmos_key})) as client:
        doc_link = '{}/docs/{}'.format(collection_link, doc_id)
        client.ReplaceItem(doc_link, new_doc)

def find_doc(cosmos_host, cosmos_key, \
        db_name, collection_name, query, \
        options=None, partition_key=None):
    db_link = 'dbs/{}'.format(db_name)
    collection_link = '{}/colls/{}'.format(db_link, collection_name)
    with IDisposable(cosmos_client.CosmosClient(cosmos_host, \
            {'masterKey': cosmos_key})) as client:
        collection = _get_or_create_collection_if_not_exist(\
            client, db_name, collection_name)
        result = list(collection.QueryItems(collection_link, query, \
            options, partition_key))
        return result

def find_doc_by_id(cosmos_host, cosmos_key, \
        db_name, collection_name, doc_id, options=None):
    db_link = 'dbs/{}'.format(db_name)
    collection_link = '{}/colls/{}'.format(db_link, collection_name)
    with IDisposable(cosmos_client.CosmosClient(cosmos_host, \
            {'masterKey': cosmos_key})) as client:
        doc_link = '{}/docs/{}'.format(collection_link, doc_id)
        return client.ReadItem(doc_link, options)

def _get_or_create_collection_if_not_exist(\
        doc_client, db_name, collection_name):
    try:
        db_link = 'dbs/{}'.format(db_name)
        doc_client.ReadDatabase(db_link)
    except cosmos_errors.HTTPFailure as e:
        if e.status_code == 404:
            doc_client.CreateDatabase({"id": db_name})
    try:
        collection_link = db_link + '/colls/{0}'.format(id)
        collection_link = '{}/colls/{}'.format(db_link, collection_name)
        doc_client.ReadContainer(collection_link)
    except cosmos_errors.HTTPFailure as e:
        if e.status_code == 404:
            doc_client.CreateContainer(db_link, {"id": collection_name})
    return doc_client
#endregion

#region HelperFunctions
def mkpath(*dirs):
    file_path = os.path.join(*dirs)
    return file_path

def mkdownloadspath(*dirs):
    file_path = os.path.join(os.path.expanduser('~'), 'Downloads')
    return os.path.join(file_path, *dirs)

def mkdir(file_path):
    try:
        os.makedirs(file_path, exist_ok=True)
    except OSError as error:
        print('Directory {} can not be created.'.format(file_path))

def _datetime_handler(x):
    if isinstance(x, datetime.datetime):
        return x.isoformat()

def _extract_dict_list_from_sql_query_result(full_query_result):
    dict_rows = []
    if full_query_result['rows']:
        for query_row in full_query_result['rows']:
            row = {}
            for i in range(len(full_query_result['description'])):
                row[full_query_result['description'][i][0]] = query_row[i]
            dict_rows.append(row)
    return dict_rows
#endregion

#region HelperClass
class IDisposable(cosmos_client.CosmosClient):
    def __init__(self, obj):
        self.obj = obj

    def __enter__(self):
        return self.obj

    def __exit__(self, exception_type, exception_val, trace):
        self = None
#endregion

if __name__ == "__main__":
    print('x')